unit rkSliderEdit;

//  SliderEdit by Roy Magne Klever
//  WEB: www.rmklever.com
//  EMail: roymagne@rmklever.com
//
//  version 1.1, March 2015


{$WARN SYMBOL_PLATFORM OFF}
{$WARN UNIT_PLATFORM OFF}

interface

uses
  Classes, Windows, SysUtils, Controls, ExtCtrls, Graphics, Forms, Messages,
  ImgList, Dialogs, Menus, StdCtrls, Math, StrUtils, Themes;

const
  CM_REFRESH = WM_USER + 1295; // Custom Message...

type
  TOnErrorEvent = procedure(Sender: TObject; AException: Byte) of object;

  TrkSliderEdit = class(TCustomControl)
  private
    { Private declarations }
    FAllowEdit: Boolean;
    FAllowKeyNav: Boolean;
    FBmp: TBitmap;
    FButton: Boolean;
    FButtonDown: Boolean;
    FButtonInc: Boolean;
    FCol1: TColor;
    FCol2: TColor;
    FCol3: TColor;
    FCol4: TColor;
    FCol5: TColor;
    FDefaultValue: Integer;
    FEditSize: Integer;
    FEditLeft: Boolean;
    FKeys: Boolean;
    FLastPoint: TPoint;
    FValue: Integer;
    FValueMin: Integer;
    FVAlueMax: Integer;
    FOnEnter: TNotifyEvent;
    FOnExit: TNotifyEvent;
    FOnMouseEnter: TNotifyEvent;
    FOnMouseLeave: TNotifyEvent;
    FOnError: TOnErrorEvent;
    FOnChange: TNotifyEvent;
    FValEdit: TEdit;
    FTransparent: Boolean;
    procedure CMEnter(var Message: TCMEnter); message CM_ENTER;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMIsToolControl(var Msg: TMessage); message CM_ISTOOLCONTROL;
    procedure WMTimer(var Msg: TWMTimer); message WM_TIMER;
    procedure WMERASEBKGND(var Message: TWMEraseBkgnd); message WM_ERASEBKGND;
    procedure WMGetDlgCode(var message: TWMGetDlgCode); message WM_GETDLGCODE;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged); message WM_WINDOWPOSCHANGED;
    procedure PaintSlider;
    procedure CMRefresh(var Message: TMessage); message CM_REFRESH;
    procedure EditExit(Sender: TObject);
    procedure EditKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure EditKeyPress(Sender: TObject; var Key: Char);
    procedure SetTransparent(const Value: Boolean);
    procedure DoKey(const Key: Word; Shift: TShiftState);
    procedure SetEditSize(const Value: Integer);
    procedure SetEditLeft(const Value: Boolean);
    procedure SetValue(const Value: Integer);
    procedure SetValueMax(const Value: Integer);
    procedure SetValueMin(const Value: Integer);
    procedure SetColorBtn(const Index: Integer; const Value: TColor);
  protected
    { Protected declarations }
    FSliderW: Integer;
    InCreate: Boolean;
    InEdit: Boolean;
    InShutdown: Boolean;
    procedure FMenuPopup(Sender: TObject);
    procedure PaintWindow(DC: HDC); override;
    procedure Resize; override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure DestroyWnd; override;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DoEdit;
    procedure ExitEditMode(SF: Boolean);
    procedure UpdateView;
  published
    { Published declarations }
    property Align;
    property AllowEdit: Boolean read FAllowEdit write FAllowEdit default True;
    property AllowKeyNav: Boolean read FAllowKeyNav write FAllowKeyNav default False;
    property Anchors;
    property Caption;
    property DefaultValue: Integer read FDefaultValue write FDefaultValue;
    property Color;
    property ColorBorder:     TColor index 0 read FCol1 write SetColorBtn default clBtnShadow;
    property ColorBrdFocused: TColor index 1 read FCol2 write SetColorBtn default clHighlight;
    property ColorGrip:       TColor index 2 read FCol3 write SetColorBtn default clBtnHighlight;
    property ColorShade1:     TColor index 3 read FCol4 write SetColorBtn default clBtnHighlight;
    property ColorShade2:     TColor index 4 read FCol5 write SetColorBtn default clBtnFace;
    property EditSize: Integer read FEditSize write SetEditSize default 32;
    //property EditLeft: Boolean read FEditLeft write SetEditLeft default False;
    property Enabled;
    property Font;
    property Height;
    property Left;
    property OnEnter: TNotifyEvent read FOnEnter write FOnEnter;
    property OnExit: TNotifyEvent read FOnExit write FOnExit;
    property OnError: TOnErrorEvent read FOnError write FOnError;
    //property ParentBackground;
    property ParentFont;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Top;
    //property Transparent: Boolean read FTransparent write SetTransparent default False;
    property Value: Integer read FValue write SetValue default 0;
    property ValueMax: Integer read FValueMax write SetValueMax default 100;
    property ValueMin: Integer read FValueMin write SetValueMin default 0;
    property Visible;
    property Width;
    property OnClick;
    property OnDblClick;
    property OnKeyDown;
    property OnKeyUp;
    property OnMouseEnter: TNotifyEvent read FOnMouseEnter write FOnMouseEnter;
    property OnMouseLeave: TNotifyEvent read FOnMouseLeave write FOnMouseLeave;
    property OnMouseDown;
    property OnMouseUp;
    property OnMouseMove;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property OnResize;
  end;

procedure Register;

implementation

const
  ArrowArea = 14;
  StepInterval = 250;

var
  StepInt: Integer;

procedure TrkSliderEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if FAllowKeyNav then DoKey(Key, Shift);
end;

procedure TrkSliderEdit.UpdateView;
begin
  PaintSlider;
end;

procedure TrkSliderEdit.ExitEditMode(SF: Boolean);
var
  p: Pointer;
begin
  if not InEdit then Exit;
  p:= FValEdit.Parent;
  FValEdit.Free;
  Invalidate;
  if (HasParent) and (SF) then SetFocus;
  InEdit:= False;
end;

procedure TrkSliderEdit.FMenuPopup(Sender: TObject);
begin

end;

procedure TrkSliderEdit.DoEdit;
var
  txt: string;
  w: Integer;
begin
  FValEdit:= TEdit.Create(nil);
  try
    InEdit:= True;
    FValEdit.Parent := Self;
    FValEdit.ParentFont := False;
    FValEdit.BorderStyle := bsNone;
    FValEdit.AutoSize := False;
    FValEdit.Ctl3D := True;
    FValEdit.Visible := False;
    FValEdit.OnKeyDown := EditKeyDown;
    FValEdit.OnKeyPress := EditKeyPress;
    FValEdit.OnExit := EditExit;
    FValEdit.Alignment:= taRightJustify;
    FValEdit.Color := Color;
    FValEdit.Left:= FSliderW + 1;
    FValEdit.Width:= FEditSize - 1;
    FValEdit.Font.Assign(Font);
    FValEdit.Height := FValEdit.Font.Size + 6;
    FValEdit.Top := ((ClientHeight - FValEdit.Height) div 2);
    FValEdit.AutoSelect := True;
    txt:= IntToStr(FValue);
    FValEdit.Text := txt;
    FValEdit.Visible := True;
    FValEdit.SetFocus;
    PaintSlider;
  except
    FValEdit.Free;
    Raise;
  end;
end;

procedure TrkSliderEdit.DoKey(const Key: Word; Shift: TShiftState);
var
  i: Integer;
begin
  if (InEdit) then Exit;
  FButtonDown := False;
  FKeys:= True;
  FLastPoint:= Point(-1, -1);
  case Key of
    VK_HOME : SetValue(FDefaultValue);
    VK_LEFT : SetValue(FValue - 1);
    VK_RIGHT: SetValue(FValue + 1);
    VK_NEXT : SetValue(FValue - 10);
    VK_PRIOR: SetValue(FValue + 10);
    VK_SPACE: DoEdit;
    else
  end;
end;

procedure TrkSliderEdit.EditKeyDown(Sender: TObject; var Key: Word; Shift:
  TShiftState);
var
  done: Boolean;
  n: Integer;
begin
  if key = VK_Escape then done := True
  else
  if key = VK_Return then begin
    done := True;
    n:= StrToIntDef(FValEdit.Text, FDefaultValue);
    SetValue(n);
    PaintSlider;
  end else
    done := False;
  if done then ExitEditMode(True);
end;

procedure TrkSliderEdit.EditKeyPress(Sender: TObject; var Key: Char);
begin
  if (Key = #13) or (Key = #27) then Key := #0;
  if not(Key in [#8, '0' .. '9', '-']) then Key := #0
  else if (Key = '-') and (Pos(Key, (Sender as TEdit).Text) > 0) then Key := #0
  else if (Key = '-') and ((Sender as TEdit).SelStart <> 0) then Key := #0;
end;

procedure TrkSliderEdit.EditExit(Sender: TObject);
begin
  ExitEditMode(False);
end;

procedure TrkSliderEdit.PaintWindow(DC: HDC);
begin
  Canvas.Lock;
  try
    Canvas.Handle := DC;
    try
      FBmp.Canvas.Font.Assign(Font);
      PaintSlider;
    finally
      Canvas.Handle := 0;
    end;
  finally
    Canvas.Unlock;
  end;
end;

procedure PreMultiplyBitmap(Bitmap: TBitmap);
var
  X, Y: Integer;
  Pixel: PRGBQuad;
begin
  with Bitmap do
    for Y := Height - 1 downto 0 do
    begin
      Pixel := ScanLine[Y];
      for X := Width - 1 downto 0 do
      begin
        Pixel.rgbReserved := 255;
        Pixel.rgbBlue := MulDiv(Pixel.rgbBlue, Pixel.rgbReserved, 255);
        Pixel.rgbGreen := MulDiv(Pixel.rgbGreen, Pixel.rgbReserved, 255);
        Pixel.rgbRed := MulDiv(Pixel.rgbRed, Pixel.rgbReserved, 255);
        inc(Pixel);
      end;
    end;
end;

procedure DrawParentImage(Control: TControl; Dest: TCanvas);
var
  SaveIndex: Integer;
  DC: HDC;
  Position: TPoint;
begin
  with Control do
  begin
    if Parent = nil then
      Exit;
    DC := Dest.Handle;
    SaveIndex := SaveDC(DC);
{$IFDEF DFS_COMPILER_2}
    GetViewportOrgEx(DC, @Position);
{$ELSE}
    GetViewportOrgEx(DC, Position);
{$ENDIF}
    SetViewportOrgEx(DC, Position.X - Left, Position.Y - Top, nil);
    IntersectClipRect(DC, 0, 0, Parent.ClientWidth, Parent.ClientHeight);
    Parent.Perform(WM_ERASEBKGND, DC, 0);
    Parent.Perform(WM_PAINT, DC, 0);
    RestoreDC(DC, SaveIndex);
  end;
end;

procedure TrkSliderEdit.PaintSlider;
const
  txtLeft = DT_END_ELLIPSIS or DT_SINGLELINE or DT_LEFT or DT_VCENTER;
  txtCenter = DT_END_ELLIPSIS or DT_SINGLELINE or DT_CENTER or DT_VCENTER;
  txtRight = DT_END_ELLIPSIS or DT_SINGLELINE or DT_RIGHT or DT_VCENTER;
var
  i, n, x, y, w, m: Integer;
  txt: string;
  R: TRect;
  ShowArrow, Done, Bool: Boolean;
begin
  if InCreate then Exit;
  Bool:= (Focused) or (InEdit);
  FSliderW:= ClientWidth - ArrowArea - FEditSize;
  FBmp.Width := ClientWidth;
  FBmp.Height := ClientHeight;
  // Base rectangle
  FBmp.Canvas.Brush.Color := ColorShade2;
  FBmp.Canvas.Brush.Style := bsSolid;
  FBmp.Canvas.Pen.Style:= psSolid;
  FBmp.Canvas.Pen.Color:= ColorBorder;
  R:= ClientRect;
  FBmp.Canvas.Rectangle(R);
  // Value rectangle
  n:= FValue - FValueMin;
  if (FValueMin < 0) and (FValueMax >= 0) then m:= Abs(FValueMin) + FValueMax else m:= FValueMax - FValueMin;
  n:= Abs(MulDiv(FSliderW, n, m));
  R:= Rect(1, 1, Max(n, 3), Height);
  FBmp.Canvas.Pen.Style:= psClear;
  FBmp.Canvas.Brush.Color:= ColorShade1;
  FBmp.Canvas.Rectangle(R);
  // Grip
  if Bool then FBmp.Canvas.Pen.Color:= ColorBrdFocused else FBmp.Canvas.Pen.Color:= ColorBorder;
  FBmp.Canvas.Pen.Style:= psSolid;
  FBmp.Canvas.Brush.Style := bsSolid;
  if FButtonDown then FBmp.Canvas.Brush.Color:= ColorGrip else FBmp.Canvas.Brush.Color:= FBmp.Canvas.Pen.Color;
  x:= Max(n, 3) - 1;
  y:= ClientHeight - 1;
  FBmp.Canvas.Polygon([Point(x - 6, y), Point(x, y - 6), Point(x + 6, y), Point(x - 6, y)]);
  // Edit rectangle
  FBmp.Canvas.Pen.Style:= psClear;
  FBmp.Canvas.Brush.Color:= Color;
  R:= Rect(FSliderW + 1, 1, ClientWidth, ClientHeight);
  FBmp.Canvas.Rectangle(R);
  // Value text
  txt:= IntToStr(FValue);
  R:= Rect(FSliderW + 1, 1, ClientWidth - ArrowArea, ClientHeight);
  FBmp.Canvas.Brush.Style := bsClear;
  DrawText(FBmp.Canvas.Handle, txt, Length(txt), R, txtRight);
  FBmp.Canvas.TextOut(2, 2, Caption);
  // Frame
  R:= ClientRect;
  FBmp.Canvas.Brush.Style:= bsClear;
  FBmp.Canvas.Pen.Style:= psSolid;
  if Bool then FBmp.Canvas.Pen.Color:= ColorBrdFocused else FBmp.Canvas.Pen.Color:= ColorBorder;
  FBmp.Canvas.Rectangle(R);
  // Arrow -
  FBmp.Canvas.Pen.Color:= clBlack;
  x:= ClientWidth - 8;
  y:= (ClientHeight shr 1) shr 1;
  FBmp.Canvas.MoveTo(x + 2, y - 1);
  FBmp.Canvas.LineTo(x, y + 1);
  FBmp.Canvas.LineTo(x + 3, y + 4);
  // Arrow +
  y:= (ClientHeight - y) - 2;
  FBmp.Canvas.MoveTo(x, y - 2);
  FBmp.Canvas.LineTo(x + 2, y);
  FBmp.Canvas.LineTo(x - 1, y + 3);
  PreMultiplyBitmap(FBmp);
  BitBlt(Canvas.Handle, 0, 0, FBmp.Width, FBmp.Height, FBmp.Canvas.Handle, 0, 0, SRCCOPY);
end;

procedure TrkSliderEdit.WMERASEBKGND(var Message: TWMEraseBkgnd);
begin
  Message.Result := 1;
end;

procedure TrkSliderEdit.WMGetDlgCode(var message: TWMGetDlgCode);
begin
  inherited;
  // Answer Delphi that this component wants to handle its own arrow key press:
  message.result := DLGC_WANTARROWS;
end;

procedure TrkSliderEdit.WMPaint(var Message: TWMPaint);
begin
  PaintHandler(Message);
end;

procedure TrkSliderEdit.WMTimer(var Msg: TWMTimer);
begin
  if IsWindowVisible(Handle) then begin
    if FButtonInc then SetValue(FValue + 1) else SetValue(FValue - 1);
    StepInt:= Trunc(StepInt / 1.1);
    if StepInt < 20 then StepInt:= 20;
    SetTimer(Handle, 0, StepInt, NIL);
  end;
  inherited;
end;

procedure TrkSliderEdit.CMEnter(var Message: TCMEnter);
begin
  inherited;
  if Assigned(OnEnter) then OnEnter(Self);
  Invalidate;
  //if FAllowKeyNav then DoKey(VK_HOME, []);
end;

procedure TrkSliderEdit.CMExit(var Message: TCMExit);
begin
  if Assigned(OnExit) then OnExit(Self);
  FKeys:= False;
  FButtonDown := False;
  Invalidate;
  inherited;
end;

procedure TrkSliderEdit.CMIsToolControl(var Msg: TMessage);
begin
  Msg.Result := 1;
end;

procedure TrkSliderEdit.WMWindowPosChanged(var Message: TWMWindowPosChanged);
begin
  inherited;
  if not (csLoading in ComponentState) then Resize;
end;

procedure TrkSliderEdit.CMRefresh(var Message: TMessage);
begin
  if (csDestroying in ComponentState) then Exit;
  FKeys:= False;
  FButtonDown:= False;
  PaintSlider;
  inherited;
end;

procedure TrkSliderEdit.SetColorBtn(const Index: Integer; const Value: TColor);
begin
  case Index of
    0 : FCol1:= Value;
    1 : FCol2:= Value;
    2 : FCol3:= Value;
    3 : FCol4:= Value;
    4 : FCol5:= Value;
    else
  end;
  Invalidate;
end;

procedure TrkSliderEdit.SetEditLeft(const Value: Boolean);
begin
  FEditLeft := Value;
  Invalidate;
end;

procedure TrkSliderEdit.SetEditSize(const Value: Integer);
begin
  FEditSize := Value;
  Invalidate;
end;

procedure TrkSliderEdit.SetTransparent(const Value: Boolean);
begin
  FTransparent := Value;
  Invalidate;
end;

procedure TrkSliderEdit.SetValue(const Value: Integer);
var i: Integer;
begin
  i:= FValue;
  FValue := Value;
  if FValue < FValueMin then FValue:= FValueMin else if FValue > FVAlueMax then FValue:= FVAlueMax;
  if Value <> i then if Assigned(FOnChange) then FOnChange(Self);
  Invalidate;
end;

procedure TrkSliderEdit.SetValueMax(const Value: Integer);
begin
  FValueMax := Value;
  if FValue < FValueMin then SetValue(FValueMin) else if FValue > FVAlueMax then SetValue(FVAlueMax);
  Invalidate;
end;

procedure TrkSliderEdit.SetValueMin(const Value: Integer);
begin
  FValueMin := Value;
  if FValue < FValueMin then SetValue(FValueMin) else if FValue > FVAlueMax then SetValue(FVAlueMax);
  Invalidate;
end;

constructor TrkSliderEdit.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  InCreate:= True;
  Width := 150;
  Height := 25;
  ControlStyle := ControlStyle + [csClickEvents, csReplicatable, csCaptureMouse];
  Color := clWindow;
  FAllowEdit:= True;
  FAllowKeyNav:= True;
  TabStop:= True;
  FBmp := TBitmap.Create;
  FBmp.PixelFormat := pf32Bit;

  FButton:= False;
  FButtonDown := False;

  Color:= clWindow;
  FCol1:= clBtnShadow;      // ColorBorder
  FCol2:= clHighlight;      // ColorBorderFocused
  FCol3:= clBtnHighlight;   // ColorGrip
  FCol4:= clBtnHighlight;   // ColorShade1
  FCol5:= clBtnFace;        // ColorShade2

  FEditSize:= 32;
  FEditLeft:= False;

  FValue:= 0;
  FValueMin:= 0;
  FValueMax:= 100;
  FTransparent:= False;
  ParentColor:= False;
  ParentFont:= True;
  InCreate := False;
  InEdit:= False;
  FSliderW:= Width - ArrowArea - FEditSize;
  FDefaultValue:= FValue;
end;

destructor TrkSliderEdit.Destroy;
begin
  InShutdown:= True;
  ExitEditMode(False);
  FBmp.Free;
  inherited Destroy;
end;

procedure TrkSliderEdit.DestroyWnd;
begin
  KillTimer(Handle, 0);
  inherited;
end;

procedure TrkSliderEdit.Resize;
begin
  if InShutDown or InCreate then Exit;
  FSliderW:= ClientWidth - ArrowArea - FEditSize;
  Invalidate;
end;

procedure TrkSliderEdit.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  m: Integer;
begin
  SetFocus;
  FButtonDown:= False;

  if Button = mbRight then begin
    SetValue(FDefaultValue);
    Invalidate;
  end else
  if (Button = mbLeft) and (x > 0) and (x < FSliderW) then
  begin
    if (FValueMin < 0) and (FValueMax >= 0) then m:= Abs(FValueMin) + FValueMax else m:= FValueMax - FValueMin;
    SetValue(MulDiv(m, x, FSliderW) + FValueMin);
    FButtonDown:= True;
    Invalidate;
    if InEdit then begin
      ExitEditMode(False);
      PaintSlider;
      Exit;
    end;
  end else
  if (x > ClientWidth - ArrowArea) and (x < ClientWidth) then begin
    if InEdit then ExitEditMode(False);
    FButtonInc:= y > ClientHeight shr 1;
    if FButtonInc then SetValue(FValue + 1) else SetValue(FValue - 1);
    StepInt:= StepInterval;
    if not (csDesigning in ComponentState) then SetTimer(Handle, 0, StepInterval, NIL);
    Invalidate;
  end else if (FAllowEdit) and (not InEdit) then DoEdit;
  inherited;
end;

procedure TrkSliderEdit.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  FButtonDown:= False;
  if not (csDesigning in ComponentState) then KillTimer(Handle, 0);
  Invalidate;
  inherited;
end;

procedure TrkSliderEdit.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  m: Integer;
begin
  if (ssLeft in Shift) and FButtonDown then begin
    if (x < 0) then x:= 0 else if (x > FSliderW) then x:= FSliderW;
    if (FValueMin < 0) and (FValueMax >= 0) then m:= Abs(FValueMin) + FValueMax else m:= FValueMax - FValueMin;
    SetValue(MulDiv(m, x, FSliderW) + FValueMin);
    PaintSlider;
  end;
  inherited;
end;

procedure TrkSliderEdit.CMMouseEnter(var Message: TMessage);
begin
  if Assigned(FOnMouseEnter) then FOnMouseEnter(Self);
  inherited
end;

procedure TrkSliderEdit.CMMouseLeave(var Message: TMessage);
begin
  FKeys:= False;
  FButtonDown := False;
  if Assigned(FOnMouseLeave) then FOnMouseLeave(Self);
  inherited;
end;

procedure Register;
begin
  RegisterComponents('rmklever', [TrkSliderEdit]);
end;

end.

